#include "machineReassignment.h"


void MachineReassignment::DependencyConstraintData::increaseCount(unsigned int neighborhood){
  if (++counts[neighborhood] == 1){
    unsigned int tail = *zero.rbegin();
    zero[nindex[neighborhood]] = tail;
    nindex[tail] = nindex[neighborhood];
    zero.pop_back();
    nindex[neighborhood] = nonzero.size();
    nonzero.push_back(neighborhood);
  }
}
void MachineReassignment::DependencyConstraintData::decreaseCount(unsigned int neighborhood){
  if (--counts[neighborhood] == 0){
    unsigned int tail = *nonzero.rbegin();  
    nonzero[nindex[neighborhood]] = tail;
    nindex[tail] = nindex[neighborhood];
    nonzero.pop_back();
    nindex[neighborhood] = zero.size();
    zero.push_back(neighborhood);
  }
}
void MachineReassignment::DependencyConstraintData::calculate(Solver *solver, Solution &solution){
  reset();
  for (vector<unsigned int>::const_iterator pite = processes->begin(); pite != processes->end(); ++pite){ 
    // *pite is id2 (resp., id) if isInGroup is true (resp., false). 
    unsigned int neighborhood;
    if (isInGroup){
      neighborhood = mr->machines[solution.assignment[*pite]].neighborhood;
      pindex[solver->processes[*pite].localId] = nprocesses[neighborhood].size();
    }
    else{
      neighborhood = mr->machines[mr->solution.assignment[*pite]].neighborhood;
    }
    increaseCount(neighborhood);
    nprocesses[neighborhood].push_back(*pite);
  }
}

MachineReassignment::int64 MachineReassignment::DependencyConstraint::calculatePenalty(Solution &solution){
  penalty = 0;

  if (service01 == inf && service02 == inf){
    for (unsigned int s = 0; s < con->mr->numServices; ++s){
      con->dependencyConstraintData[s].calculate(con->solver, solution);
    }
  }
  else{
    DependencyConstraintData &data1 = con->dependencyConstraintData[service01];
    DependencyConstraintData &data2 = con->dependencyConstraintData[service02];
    for (vector<unsigned int>::iterator nite = data2.zero.begin(); nite != data2.zero.end(); ++nite){
      penalty += data1.nprocesses[*nite].size(); 
    }
  }
  return penalty;
}

void MachineReassignment::DependencyConstraint::update(unsigned int process, unsigned int neighborhood0, unsigned int neighborhood){
  // neighborhood != neighborhood is asumed
  unsigned int service = con->solver->services[con->solver->processes[process].service].id;
  DependencyConstraintData &data = con->dependencyConstraintData[service];
  unsigned int index = data.pindex[con->solver->processes[process].localId];
  vector<vector<unsigned int> > &nprocesses = data.nprocesses;
  unsigned int tail = *nprocesses[neighborhood0].rbegin();
  nprocesses[neighborhood0][index] = tail;
  data.pindex[con->solver->processes[tail].localId] = index;
  nprocesses[neighborhood0].pop_back();
  data.pindex[con->solver->processes[process].localId] = nprocesses[neighborhood].size();
  nprocesses[neighborhood].push_back(process);

  data.decreaseCount(neighborhood0);
  data.increaseCount(neighborhood);
}

MachineReassignment::int64 MachineReassignment::DependencyConstraint::updatePenalty(Solution &solution, Solver::Move *move){
  ppenalty = penalty;
  switch(move->type){
  case Solver::Shift:
    {
      Solver::ShiftMove *shift = dynamic_cast<Solver::ShiftMove*>(move);
      unsigned int neighborhood0 = con->mr->machines[solution.assignment[shift->process]].neighborhood;
      unsigned int neighborhood = con->mr->machines[shift->machine].neighborhood;
      if (neighborhood != neighborhood0){
        if (service01 == inf && service02 == inf){
          update(shift->process, neighborhood0, neighborhood);
        }
        else{
          DependencyConstraintData &data2 = con->dependencyConstraintData[service02];
          if (con->solver->processes[shift->process].service == service1){
              if (data2.counts[neighborhood0] == 0) --penalty;
              if (data2.counts[neighborhood] == 0) ++penalty;
            skip = (penalty == ppenalty);
          }
          else if (con->solver->processes[shift->process].service == service2){
            DependencyConstraintData &data1 = con->dependencyConstraintData[service01];
            skip = true;
            if (data2.counts[neighborhood0] == 0){
              penalty += data1.nprocesses[neighborhood0].size();
              skip = false;
            }
            if (data2.counts[neighborhood] == 1){
              penalty -= data1.nprocesses[neighborhood].size();
              skip = false;
            }
          }
        }
      }
      break;
    }
  case Solver::Swap:
    {
      Solver::SwapMove *swap = static_cast<Solver::SwapMove*>(move);
      unsigned int neighborhood1 = con->mr->machines[solution.assignment[swap->process1]].neighborhood;
      unsigned int neighborhood2 = con->mr->machines[solution.assignment[swap->process2]].neighborhood;
      if (neighborhood1 != neighborhood2){
        if (service01 == inf && service02 == inf){
          update(swap->process1, neighborhood1, neighborhood2);
          update(swap->process2, neighborhood2, neighborhood1);
        }
        else{
          DependencyConstraintData &data2 = con->dependencyConstraintData[service02];
          if (con->solver->processes[swap->process1].service == service1){ // process1 and process2 belong to the same service 
            skip = ((data2.counts[neighborhood1] == 0) != (data2.counts[neighborhood2] == 0));
          }
          else if (con->solver->processes[swap->process1].service == service2){
            skip = (data2.counts[neighborhood1] <= 1 || data2.counts[neighborhood2] <= 1);
          }
        }
      }
      break;
    }
  }
  return penalty - ppenalty;
}

void MachineReassignment::DependencyConstraint::calculateDiff(Solution &solution){
  if (service01 == inf && service02 == inf){
    return;
  }
  // calculate diff
  vector<unsigned int>::const_iterator pite;
  vector<unsigned int>::iterator nite, mite;
  // move process in service1
  DependencyConstraintData &data1 = con->dependencyConstraintData[service01];
  DependencyConstraintData &data2 = con->dependencyConstraintData[service02];
  if (service1 != inf){
    for (pite = data1.processes->begin(); pite != data1.processes->end(); ++pite){
      unsigned int neighborhood = con->mr->machines[solution.assignment[*pite]].neighborhood;
      if (data1.counts[neighborhood] == 0){
        for (nite = data2.nonzero.begin(); nite != data2.nonzero.end(); ++nite){
          setAdjustmentForNeighborhood(*pite, *nite, -1);
        }      
      }
      else{
        for (nite = data2.zero.begin(); nite != data2.zero.end(); ++nite){
          setAdjustmentForNeighborhood(*pite, *nite, 1);
        }      
      }
    }
  }
  // move process in service2
  if (service2 != inf){
    for (pite = data2.processes->begin(); pite != data2.processes->end(); ++pite){
      unsigned int neighborhood = con->mr->machines[solution.assignment[*pite]].neighborhood;
      int64 diff = 0;
      if (data2.counts[neighborhood] == 1){
        diff = data1.nprocesses[neighborhood].size();
      }
      for (unsigned int n = 0; n < con->mr->numNeighborhoods; ++n){
        if (n != neighborhood){
          int64 diff2 = diff;
          if (data2.counts[n] == 0) diff2 = diff - data1.nprocesses[n].size();
          setAdjustmentForNeighborhood(*pite, n, diff2);
        }
      }
    }
  }
}


void MachineReassignment::DependencyConstraint::prepare(MachineReassignment::Solver *solver){
  MachineReassignment *mr = con->mr;
  con->addConstraint(this);
  
  if (service01 == inf && service02 == inf){ // for dummy constraint which is to update DependencyConstraintData
    for (unsigned int p = 0; p < solver->numProcesses; ++p){
      involve(p, inf);
    }
    return;
  }


  vector<unsigned int>::const_iterator ite;
  if (con->dependencyConstraintData[service01].isInGroup){
    service1 = mr->services[service01].id2;
    vector<unsigned int> *processes1 = con->dependencyConstraintData[service01].processes;
    for (ite = processes1->begin(); ite != processes1->end(); ++ite){
      involve(*ite,inf);
    }
  }
  else{
    service1 = inf;
  }
  if (con->dependencyConstraintData[service02].isInGroup){
    service2 = mr->services[service02].id2;
    vector<unsigned int> *processes2 = con->dependencyConstraintData[service02].processes;
    for (ite = processes2->begin(); ite != processes2->end(); ++ite){
      involve(*ite, inf);
    }
  }
  else{
    service2 = inf;
  }
}

// ----- Dependency Constraints -----
void MachineReassignment::Constraints::initDependencyConstraints(){
  dependencyConstraintData.resize(mr->numServices);
  for (unsigned int s = 0; s < mr->numServices; ++s){
    dependencyConstraintData[s].init(mr, s);
  }

  dependencyConstraintForPreparation = new DependencyConstraint(this, inf, inf, inf); 
  dependencyConstraints.resize(mr->numServices);
  for (unsigned int s1 = 0; s1 < mr->numServices; ++s1){
    vector<unsigned int> &dependencies = mr->services[s1].dependencies;
    for (vector<unsigned int>::iterator ite = dependencies.begin(); ite != dependencies.end(); ++ite){
      if (s1 != *ite){
        dependencyConstraints[s1].push_back(new DependencyConstraint(this, inf, s1, *ite));
      }
    }
  }
}
void MachineReassignment::Constraints::prepareDependencyConstraints(){
  for (unsigned int s = 0; s < mr->numServices; ++s){
    dependencyConstraintData[s].prepare(solver);
  }

  dependencyConstraintForPreparation->prepare(solver);
  for (unsigned int s1 = 0; s1 < mr->numServices; ++s1){
    vector<unsigned int> &dependencies = mr->services[s1].dependencies;
    for (unsigned int dd = 0; dd < dependencies.size(); ++dd){
      unsigned int s2 = dependencies[dd];
      if (s1 != s2 && (dependencyConstraintData[s1].isInGroup || dependencyConstraintData[s2].isInGroup)){
        dependencyConstraints[s1][dd]->prepare(solver);
      }
    }
  }
}
