#include <iostream>
#include <string>
#include <vector>
#include <set>
#include <fstream>
#include <stdexcept>
#include <climits>  
#include <cstdlib>
#include <cstring>
#include <cfloat>

#include "cputime.h"
#include "random.h"

#pragma warning (disable : 4267)
#pragma warning (disable : 4996)
#define SILENT   1

using namespace std;

class MachineReassignment{
public:
  typedef long long int64;

  MachineReassignment(): invalid(true), maxCpuTime(300) {}
  MachineReassignment(const MachineReassignment&);
  MachineReassignment &operator=(const MachineReassignment&);
  ~MachineReassignment() {}

  void setFileNames(string modelFileName, string assignmentFileName, string solutionFileName){ 
    this->modelFileName = modelFileName; 
    this->assignmentFileName = assignmentFileName; 
    this->solutionFileName = solutionFileName; 
  }
  void readProblem();
  void setMaxCpuTime(double t){ maxCpuTime = t; }
  void setRandomSeed(int s){ rand.init(s); }
  void solve();
  int64 getPenalty(){ 
    return (invalid || solution.penalty.hard > 0)? LLONG_MAX : solution.penalty.soft;
  }
  void outputSolution();

private:
  static const int64 inf = static_cast<int64>(INT_MAX);
  static const unsigned int MaxNumMachinesParallel;
  
  // ----- read number -----
  unsigned int readNumber(char*&);

  // ----- resource -----
  struct Resource{
    bool transient;
    unsigned int weightLoadCost;
  };
  // ----- machine -----
  struct Machine{
    unsigned int neighborhood;
    unsigned int location;
    vector<unsigned int> capacities, saftyCapacities; // for each resource
    vector<unsigned int> movingCosts; // for each machine;
  };
  // ----- service -----
  struct Service{
    unsigned int id, id2, group;
    unsigned int spreadMin;
    unsigned int numDependencies;
    vector<unsigned int> dependencies;
    vector<unsigned int> processes;
  };
  // ----- process -----
  struct Process{
    unsigned int id, localId, id2;
    unsigned int service;
    vector<unsigned int> requirements; // for each resource
    unsigned int movingCost;
    unsigned int originalAssignment;
  };
  // ----- balance -----
  struct Balance{
    unsigned int resource1, resource2;
    unsigned int target;
    unsigned int weight;
  };

  // ----- instance data -----
  unsigned int numResources;
  unsigned int numMachines;
  unsigned int numNeighborhoods;
  unsigned int numLocations;
  unsigned int numServices;
  unsigned int numProcesses;
  unsigned int numBalances;
  vector<Resource> resources;
  vector<Machine> machines;
  vector<vector<unsigned int> > machinesInNeighborhoods;
  vector<vector<unsigned int> > machinesInLocation;
  vector<Service> services;
  vector<Process> processes;
  vector<Balance> balances;
  //   weights
  unsigned int weightProcessMovingCost;
  unsigned int weightServiceMovingCost;
  unsigned int weightMachineMovingCost;

  // ----- penalty -----
  struct Penalty{
    Penalty(int64 hard_ = 0, int64 soft_ = 0): hard(hard_), soft(soft_){}
    int64 hard;
    int64 soft;
    bool operator < (const Penalty &x) const { return hard < x.hard || hard == x.hard && soft < x.soft; }
    bool operator <= (const Penalty &x) const { return hard < x.hard || hard == x.hard && soft <= x.soft; }
    bool operator >= (const Penalty &x) const { return hard > x.hard || hard == x.hard && soft >= x.soft; }
    bool operator == (const Penalty &x) const { return hard == x.hard && soft == x.soft; }
    bool operator == (int64 x) const { return hard * inf + soft == x; }
  };
  // ----- Solution -----
  struct Solution{
    Solution &operator =(const Solution &solution){
      assignment = solution.assignment;
      penalty = solution.penalty;
      eval = solution.eval;
      return *this;
    }
    void initialize(unsigned int numProcesses){ 
      assignment.resize(numProcesses); 
      penalty.hard = inf;
    }
    void clearPenalty(){
      penalty = 0;
      eval = 0;
    }
    vector<unsigned int> assignment;
    Penalty penalty;
    double eval;
  };


  struct Constraints;
  // ----- solver -----
  class Solver{
    friend class MachineReassignment;
  public:
    Solver(MachineReassignment *mr_, Constraints *con_, vector<Service> &s, vector<Process> &p): 
        mr(mr_), con(con_), services(s), processes(p), cputime(mr->cputime), maxCpuTime(mr->maxCpuTime), iteration(mr->iteration){
        numServices = services.size();
        numProcesses = processes.size();
        rand.init(mr->rand(0,inf)); 
    }
    Solver(const Solver&);
    Solver &operator=(const Solver&);
    ~Solver(){}

    bool solve(); // return true when a stopping criterion is met.

  private:
    static const double dinf;
    static const double eps;
    static const double ratioResetWeights;
    static const unsigned int maxNumRounds;
    enum MoveType { Shift, Swap, Adjustment };  

    // ----- Move -----
    struct Move{
      Move(){}
      void clear(){ 
        diffEval = 0;
      }
      enum MoveType type;
      double diffEval; // difference of the weighted penalty (i.e., evaluation function)
      virtual void apply(Solution &solution) = 0;
    };
    struct ShiftMove: public Move{
      ShiftMove(){ type = Shift; }
      ShiftMove(unsigned int p, unsigned int m): process(p), machine(m){ type = Shift; }
      unsigned int process, machine;
      unsigned int machine0;
      void apply(Solution &solution){
        machine0 = solution.assignment[process];
        solution.assignment[process] = machine;
      }
    };
    struct SwapMove: public Move{
      SwapMove(){ type = Swap; }
      unsigned int process1, process2;
      void apply(Solution &solution){
        swap(solution.assignment[process1], solution.assignment[process2]);
      }
    };
    struct EvalAdjustment: public Move{
      EvalAdjustment(){ type = Adjustment; }
      unsigned int id;
      void apply(Solution &solution){
         throw runtime_error("it is impossible to apply EvalAdjustment");
      }
    };

    // ----- DiffPenalty-----
    struct DiffPenalty{
      DiffPenalty(Move *m, int64 d): move(m), diff(d){}
      Move *move;
      int64 diff;
    };

    // ----- Moves -----        
    class Moves{
    public:
      Moves(){}
      void initializeMoves(Solver *solver);
      void clear(){ 
        for (unsigned int p = 0; p < shiftMoves.size(); ++p){
          for (vector<ShiftMove>::iterator ite = shiftMoves[p].begin(); ite != shiftMoves[p].end(); ++ite){
            ite->clear();
          }
        }
        for (unsigned int s = 0; s < swapMoves.size(); ++s){
          for (vector<SwapMove>::iterator ite = swapMoves[s].begin(); ite != swapMoves[s].end(); ++ite){
            ite->clear();
          }
        }
        for (vector<Move*>::iterator ite = adjustments.begin(); ite != adjustments.end(); ++ite){
          (*ite)->clear();
        }
      }
      vector<vector<ShiftMove> > shiftMoves;
      vector<vector<SwapMove> > swapMoves;
      vector<vector<SwapMove*> > swapMovesByProcess; 
      vector<Move*> adjustments;
      vector<EvalAdjustment> adjustmentsForProcess; 
      vector<vector<EvalAdjustment> > adjustmentsForNeighborhood, adjustmentsForLocation; 
    } moves;


    // ----- TabuList -----
    class TabuList{
    public:
      TabuList(): tenure(1){}
      void initialize(unsigned int size){
        counters.resize(size, 0);
        maxCounter = 0;
        counter = 1;
      }
      void setTenure(unsigned int t){ tenure = t; }
      void clear(){ counter = maxCounter+1; }
      void increment(){ ++counter; }
      void setTabu(unsigned int id){ setTabu(id, tenure); }
      void setTabu(unsigned int id, unsigned int t){ 
        counters[id] = counter + t; 
        maxCounter = max(maxCounter, counters[id]);
      }
      bool isTabu(unsigned int id){ return counter <= counters[id]; }
    private:
      unsigned int counter;
      unsigned int tenure;
      unsigned int maxCounter;
      vector<unsigned int> counters;
    };

    // ----- method -----
    void initialize(Solution &);
    bool localSearch();
    void applyMove(Solution&, Move*);
    void updateIncumbent(Solution&);
    void kick();

    bool reassignService(Solution&);

    // ----- member -----
    MachineReassignment *mr;
    Constraints *con;
    unsigned int numServices;
    unsigned int numProcesses;
    vector<Service> &services;
    vector<Process> &processes;
    
    CpuTime &cputime;
    double maxCpuTime;
    Random rand;
    bool adjustWeight, bestAdmissible, useKick, oneGroup;

    Solution current, localBest, incumbent;

    unsigned int &iteration;
    TabuList tabulist;
    Penalty prevBestPenalty; 
    unsigned int numRuns, numRounds;
  };

  // ----- Constraint -----
  class Constraint{
  public:
    Constraint(Constraints *con_, int64 w): con(con_), weight0(w){
      id0 = con->constraints0.size();
      con->constraints0.push_back(this);
      weight = static_cast<double>(weight0);
      skip = false;
    }
    virtual ~Constraint(){}
    virtual void prepare(Solver *solver) = 0;
    virtual void clear(){
      vector<Solver::DiffPenalty>().swap(diffs);
    };
    virtual int64 calculatePenalty(Solution&) = 0;
    virtual int64 updatePenalty(Solution&, Solver::Move*) = 0; // return difference of penalty; i.e., penalty-ppenalty
    virtual void calculateDiff(Solution&) = 0;

    unsigned int id0, id;
    int64 penalty, ppenalty; // ppenalty (previous penalty); used in updatePenaltyDiff
    int64 weight0; // original weight
    double weight; // not larger than weight0; used to define the evaluation function; 
    vector<Solver::DiffPenalty> diffs; // previous and current differences
    bool skip;
  protected:
    void involve(unsigned int process, unsigned int machine){
      con->involve(id, process, machine);
    }
    void setShiftDiffPenalty(unsigned int process, unsigned int machine, int64 diff){ 
      if (diff == 0) return;
      diffs.push_back(Solver::DiffPenalty(&con->solver->moves.shiftMoves[process][machine], diff));
    }
    void setSwapDiffPenalty(Solver::SwapMove *move, int64 diff){ 
      if (diff == 0) return;
      diffs.push_back(Solver::DiffPenalty(move, diff));
    }
    void setSwapDiffPenalty(unsigned int service, unsigned int localId, int64 diff){
      setSwapDiffPenalty(&con->solver->moves.swapMoves[service][localId], diff);
    }
    void setAdjustmentForProcess(unsigned int process, int64 diff){ 
      if (diff == 0) return;
      diffs.push_back(Solver::DiffPenalty(&con->solver->moves.adjustmentsForProcess[process], diff));
    }
    void setAdjustmentForNeighborhood(unsigned int process, unsigned int neighborhood, int64 diff){ 
      if (diff == 0) return;
      diffs.push_back(Solver::DiffPenalty(&con->solver->moves.adjustmentsForNeighborhood[process][neighborhood], diff));
    }
    void setAdjustmentForLocation(unsigned int process, unsigned int location, int64 diff){ 
      if (diff == 0) return;
      diffs.push_back(Solver::DiffPenalty(&con->solver->moves.adjustmentsForLocation[process][location], diff));
    }
    static int64 func(int64 value){ return (value > 0)? value : 0; }
    Constraints *con;
  };
  // ----- ResourceConstraints  -----
  class ResourceConstraint: public Constraint{
  public:
    ResourceConstraint(Constraints *con_, int64 weight, unsigned int m): Constraint(con_, weight), machine(m), total0(0){
      con = con_;
    }
    virtual ~ResourceConstraint(){}
    void prepare(Solver *solver);
    void prepare2(Solver *solver);
    void postprocess(Solution&);
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
    int64 calculateDiffSwitch(unsigned int process_out, unsigned int process_in){
      return func(total+requirement(process_in)-requirement(process_out)-capacity) - penalty;
    }
    unsigned int machine;
    int64 capacity, capacity0, total0;
  protected:
    virtual int64 requirement(unsigned int) = 0;
    Constraints *con;
    int64 total, totalBackup;
    int64 maxRequirement;
  };
  //
  class CapacityConstraint: public ResourceConstraint{
  public:
    CapacityConstraint(Constraints *con, int64 weight, unsigned int m, unsigned int r)
      : ResourceConstraint(con, weight, m), resource(r){}
    int64 requirement(unsigned int process){
      if (con->mr->resources[resource].transient && con->solver->processes[process].originalAssignment == machine){
        return 0;
      }  
      else{
        return con->solver->processes[process].requirements[resource];
      }

    }
  private:
    unsigned int resource;
  };
  //
  class LoadCost: public ResourceConstraint{
  public:
    LoadCost(Constraints *con, int64 weight, unsigned int m, unsigned int r)
      : ResourceConstraint(con, weight, m), resource(r){}
    int64 requirement(unsigned int process){
      return con->solver->processes[process].requirements[resource];
    }
  private:
    unsigned int resource;
  };
  //
  class BalanceCost: public ResourceConstraint{
  public:
    BalanceCost(Constraints *con, int64 weight, unsigned int m, unsigned int b, unsigned int r1, unsigned int r2)
      : ResourceConstraint(con, weight, m), balance(b), resource1(r1), resource2(r2){}
    int64 requirement(unsigned int process){
      return -1 * static_cast<int64>(con->mr->balances[balance].target)*con->solver->processes[process].requirements[resource1] 
        + con->solver->processes[process].requirements[resource2];
    }
  private:
    unsigned int balance;
    unsigned int resource1, resource2;
  };

  // ----- ConflictConstraints -----
  class ConflictConstraint: public Constraint{
  public:
    ConflictConstraint(Constraints *con_, int64 weight, unsigned int s): Constraint(con_, weight), service0(s){}
    void clear(){
      vector<Solver::DiffPenalty>().swap(diffs);
      vector<unsigned int>().swap(counts);
      vector<unsigned int>().swap(zero);
      vector<unsigned int>().swap(nonzero);
      vector<unsigned int>().swap(index);
    };
    void prepare(Solver *solver){
      con->addConstraint(this);
      counts.resize(solver->mr->numMachines);
      index.resize(solver->mr->numMachines);
      service = solver->mr->services[service0].id2;
      processes = &solver->services[service].processes;
      for (vector<unsigned int>::const_iterator ite = processes->begin(); ite != processes->end(); ++ite){
        involve(*ite, inf);
      }
    }
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
  
    unsigned int service0, service;
  private:
    const vector<unsigned int> *processes;
    vector<unsigned int> counts;
    vector<unsigned int> zero, nonzero;
    vector<unsigned int> index;
    void reset();
    void increaseCount(unsigned int m);
    void decreaseCount(unsigned int m);
  };
  // ----- SpreadConstraints -----
  class SpreadConstraint: public Constraint{
  public:
    SpreadConstraint(Constraints *con_, int64 weight, unsigned int s): Constraint(con_, weight), service0(s){}
    void clear(){
      vector<Solver::DiffPenalty>().swap(diffs);
      vector<unsigned int>().swap(counts);
      vector<unsigned int>().swap(zero);
      vector<unsigned int>().swap(nonzero);
      vector<unsigned int>().swap(index);
    };
    void prepare(Solver *solver);
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
  
    unsigned int service0, service;
  private:
    const vector<unsigned int> *processes;
    vector<unsigned int> counts;
    vector<unsigned int> zero, nonzero;
    vector<unsigned int> index;
    void reset();
    void increaseCount(unsigned int l);
    void decreaseCount(unsigned int l);
    void update(unsigned int process, unsigned int machine0, unsigned int machine);
    int64 spreadMin;
  };
  // ----- DependencyConstraints -----
  struct DependencyConstraintData{
    void init(MachineReassignment *mr_, unsigned int s){
      mr = mr_;
      service = s;
      counts.resize(mr->numNeighborhoods);
      nindex.resize(mr->numNeighborhoods);
      nprocesses.resize(mr->numNeighborhoods);
      pindex.resize(mr->services[service].processes.size());
    }
    void prepare(Solver *solver){
      isInGroup = (mr->services[service].group == solver->con->group);
      if (isInGroup){
        processes = &solver->services[mr->services[service].id2].processes;
      }
      else{
        processes = &mr->services[service].processes;
      }
    }
    void reset(){
      nonzero.clear();
      zero.clear();
      for (unsigned int n = 0; n < mr->numNeighborhoods; ++n){
        nprocesses[n].clear();
        counts[n] = 0;
        nindex[n] = n;
        zero.push_back(n);
      }
    }
    void increaseCount(unsigned int neighborhood);
    void decreaseCount(unsigned int neighborhood);

    void calculate(Solver*, Solution&);
    MachineReassignment *mr;
    unsigned int service;
    bool isInGroup;
    vector<unsigned int> *processes;

    vector<unsigned int> counts;
    vector<unsigned int> zero, nonzero;
    vector<unsigned int> nindex;

    vector<vector<unsigned int> > nprocesses; // nprocesses[N]: a set of processes assigned a machine in N
    vector<unsigned int> pindex;
  };
  class DependencyConstraint: public Constraint{
  public:
    DependencyConstraint(Constraints *con_, int64 weight, unsigned int s1, unsigned int s2): Constraint(con_, weight), service01(s1), service02(s2){}
    void prepare(Solver *solver);
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
  
    unsigned int service01, service02, service1, service2;
  private:
    void update(unsigned int process, unsigned int neighborhood0, unsigned int neighborhood);
  };
   // ----- ProcessMachineMovingCost -----
  class ProcessMachineMovingCost: public Constraint{
  public:
    ProcessMachineMovingCost(Constraints *con_, int64 weight, unsigned int p): Constraint(con_, weight), process0(p){}
    void clear(){
      vector<Solver::DiffPenalty>().swap(diffs);
      vector<int64>().swap(costs);
    };
    void prepare(Solver *solver);
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
    int64 calculateDiffSwitch(unsigned int machine_out, unsigned int machine_in){
      return costs[machine_in] - costs[machine_out];
    }
    unsigned int process0, process;
    vector<int64> costs;
  private:
  };
  // ----- ServiceMovingCost -----
  class ServiceMovingCost: public Constraint{
  public:
    ServiceMovingCost(Constraints *con_, int64 weight): Constraint(con_, weight){
      ub = 0;
      for (unsigned int s = 0; s < con->mr->numServices; ++s){
        ub = max<unsigned int>(ub, con->mr->services[s].processes.size());
      }
      services.resize(ub+1);
      cindex.resize(ub+1);
      sindex.resize(con->mr->numServices);
      counts.resize(con->mr->numServices);
    }
    void prepare(Solver *solver);
    int64 calculatePenalty(Solution&); // called in solve()
    int64 updatePenalty(Solution&, Solver::Move*);
    void calculateDiff(Solution&); 
  
  private:
    unsigned int ub;
    vector<vector<unsigned int> > services; // count -> a set of services
    vector<set<int64,greater<int64> >::iterator> cindex;
    set<int64,greater<int64> > tree;
    vector<unsigned int> sindex;
    vector<unsigned int> counts;
    void makeListsAndTree();
    void changeCount(unsigned int service, int diff);
    int swapDiff(Solution &solution, Solver::SwapMove *swap);
  };
  
  // ----- Constraints -----
  struct Constraints{
    Constraints(MachineReassignment *mr_): mr(mr_){}
    ~Constraints(){
      for (unsigned int c = 0; c < constraints0.size(); c++){
			  delete constraints0[c];
		  }
    }
    void initialize(){
      if (mr->solution.assignment.size() != mr->numProcesses){
        throw runtime_error("Constraints::initialize() must be called with 'solution' properly initialized");
      }
      initCapacityConstraints();
      initConflictConstraints();
      initSpreadConstraints();
      initDependencyConstraints();
      initLoadCosts();
      initBalanceCosts();
      initProcessMachineMovingCosts();
      initServiceMovingCost();
      processMachineMovingCostsTotalBackup = 0;
    }
    void involve(unsigned int constraint, unsigned int process, unsigned int machine){
      if (machine == inf){
        (*constraintIds2[process].rbegin()) = constraint;
        constraintIds2[process].push_back(INT_MAX); // sentinel
      }
      else if (process == inf){
        (*constraintIds3[machine].rbegin()) = constraint;
        constraintIds3[machine].push_back(INT_MAX); // sentinel
      }
      else{
        (*constraintIds[process][machine].rbegin()) = constraint;
        constraintIds[process][machine].push_back(INT_MAX); // sentinel
      }
    }
    void addConstraint(Constraint *constraint){
      constraint->id = constraints.size();
      constraints.push_back(constraint);
    }
    void prepare(Solver *solver_){
      solver = solver_;
      group = solver->services[0].group;
      for (vector<Constraint*>::iterator ite = constraints.begin(); ite != constraints.end(); ++ite){
        (*ite)->Constraint::clear();
        (*ite)->clear();
      }
      constraints.clear();
      constraintIds.resize(solver->numProcesses);
      constraintIds2.resize(solver->numProcesses);
      constraintIds3.resize(mr->numMachines);
      for (unsigned int p = 0; p < solver->numProcesses; ++p){
        constraintIds[p].resize(mr->numMachines);
        for (unsigned int m = 0; m < mr->numMachines; ++m){
          constraintIds[p][m].clear();
          constraintIds[p][m].push_back(INT_MAX);
        }
        constraintIds2[p].clear();
        constraintIds2[p].push_back(INT_MAX);
      }
      for (unsigned int m = 0; m < mr->numMachines; ++m){
        constraintIds3[m].clear();
        constraintIds3[m].push_back(INT_MAX);
      }
      prepareCapacityConstraints();
      prepareConflictConstraints();
      prepareSpreadConstraints();
      prepareDependencyConstraints();
      prepareLoadCosts();
      prepareBalanceCosts();
      prepareProcessMachineMovingCosts();
      prepareServiceMovingCost();
    }

    void postprocess(Solution &solution){
      postprocessCapacityConstraints(solution);
      postprocessLoadCosts(solution);
      postprocessBalanceCosts(solution);
      postprocessProcessMachineMovingCosts(solution);

      vector<vector<vector<unsigned int> > >().swap(constraintIds);
      vector<vector<unsigned int> >().swap(constraintIds2);
      vector<vector<unsigned int> >().swap(constraintIds3);
    }

    int64 getConstant(){
      return processMachineMovingCostsTotal0 - processMachineMovingCostsTotalBackup;
    }

    void initCapacityConstraints();
    void initConflictConstraints();
    void initSpreadConstraints();
    void initDependencyConstraints();
    void initLoadCosts();
    void initBalanceCosts();
    void initProcessMachineMovingCosts();
    void initServiceMovingCost();

    void prepareCapacityConstraints();
    void prepareConflictConstraints();
    void prepareSpreadConstraints();
    void prepareDependencyConstraints();
    void prepareLoadCosts();
    void prepareBalanceCosts();
    void prepareProcessMachineMovingCosts();
    void prepareServiceMovingCost();

    void postprocessCapacityConstraints(Solution&);
    void postprocessLoadCosts(Solution&);
    void postprocessBalanceCosts(Solution&);
    void postprocessProcessMachineMovingCosts(Solution&);

    // ----- members -----
    MachineReassignment *mr;
    Solver *solver;
    unsigned int group;
    vector<DependencyConstraintData> dependencyConstraintData;
    DependencyConstraint *dependencyConstraintForPreparation;
    int64 processMachineMovingCostsTotal0, processMachineMovingCostsTotalBackup;

    vector<Constraint*> constraints0, constraints;
    vector<vector<vector<unsigned int> > > constraintIds; // (process,machine) -> set of constraints
    vector<vector<unsigned int> > constraintIds2; // process -> set of constraints
    vector<vector<unsigned int> > constraintIds3; // machine -> set of constraints

    vector<vector<ResourceConstraint*> > capacityConstraints, loadCosts, balanceCosts;
    vector<ConflictConstraint*> conflictConstraints;
    vector<SpreadConstraint*> spreadConstraints;
    vector<vector<DependencyConstraint*> > dependencyConstraints;
    vector<ProcessMachineMovingCost*> processMachineMovingCosts;
    ServiceMovingCost* serviceMovingCost;
  };

  // members
  bool invalid;
  Solution solution;
  CpuTime cputime;
  Random rand;
  unsigned int iteration;

  //   parameters
  string modelFileName, assignmentFileName, solutionFileName;
  double maxCpuTime;
};
