#include "machineReassignment.h"
#include <hash_set>

#include "matching_descent.h"

const double MachineReassignment::Solver::dinf = DBL_MAX;
const double MachineReassignment::Solver::eps = 1e-6;
const double MachineReassignment::Solver::ratioResetWeights = 0.6;
const unsigned int MachineReassignment::Solver::maxNumRounds = 2;

// ----- Solve -----
void MachineReassignment::solve(){
  unsigned int p, s, g, pp, ss;
  unsigned int size = min<unsigned int>(10000,1000000/numMachines);  
  vector<vector<Service> > services2;
  vector<vector<Process> > processes2;
  services2.resize(numProcesses/size+1);
  processes2.resize(numProcesses/size+1);

  for (g = 0, s = 0; s < numServices; ++s){
    services[s].group = g;
    services[s].id2 = services2[g].size();
    services2[g].push_back(services[s]);
    for (pp = 0; pp < services[s].processes.size(); ++pp){
      p = services[s].processes[pp];
      processes[p].id2 = processes2[g].size();
      processes2[g].push_back(processes[p]);
    }
    if (processes2[g].size() >= size){
      ++g;
    }
  }
  while (services2.rbegin()->empty()){
    services2.pop_back();
    processes2.pop_back();
  }
  unsigned int numGroups = services2.size();
  for (g = 0; g < numGroups; ++g){
    for (ss = 0; ss < services2[g].size(); ++ss){
      Service &service = services2[g][ss];
      for (pp = 0; pp < service.processes.size(); ++pp){
        service.processes[pp] = processes[service.processes[pp]].id2;
      }
    }
    for (pp = 0; pp < processes2[g].size(); ++pp){
      processes2[g][pp].service = services[processes2[g][pp].service].id2;
    }
  }
  // ----- initialize solution -----
  int cpulimitMD = (numMachines <= 200)? 0 : 20;

  solution.initialize(numProcesses);
  solution.penalty.hard = inf;
  invalid = false;

  if (cpulimitMD > 0){
    MatchingDescent *matchingDescent = new MatchingDescent;
    matchingDescent->solve(modelFileName.c_str(), assignmentFileName.c_str(), cpulimitMD);  
    for (p = 0; p < numProcesses; ++p){
      solution.assignment[p] = matchingDescent->getX(p);
    }
    delete matchingDescent;
  }
  else{
    for (p = 0; p < numProcesses; ++p){
      solution.assignment[p] = processes[p].originalAssignment;
    }
  }

  // ----- solve -----
  Constraints constraints(this);
  constraints.initialize();
  Solver *solver;
  iteration = 0;
  for (unsigned int round = 0, group = 0; ; ++round, ++group){
    if (group >= numGroups){
      group = 0;
    }
#if !SILENT
    cout << "# round = " << round << ", group = " << group+1 << "/" << numGroups << endl;
#endif
    bool ret;
    try {
      solver = new Solver(this, &constraints, services2[group], processes2[group]);
      solver->oneGroup = (numGroups == 1)? true : false;
      solver->adjustWeight = ((numMachines <= 100) && (round >= numGroups || solver->oneGroup))? true : false;
      solver->bestAdmissible = (cpulimitMD == 0)? true: false;
      solver->useKick = (round >= numGroups*2)? true: false;
      ret = solver->solve();
    }
    catch (bad_alloc){
      cout << "bad alloc" << endl;
    }
    if (solver->incumbent.penalty < solution.penalty){
      solution.penalty = solver->incumbent.penalty;
      for (pp = 0; pp < processes2[group].size(); ++pp){
        solution.assignment[processes2[group][pp].id] = solver->incumbent.assignment[pp];
      }
      constraints.postprocess(solver->incumbent);
    }
    delete solver;
    
    if (ret) break;
  } 
}


// ----- Initialize Moves -----
void MachineReassignment::Solver::Moves::initializeMoves(Solver *solver){
  // shift
  shiftMoves.resize(solver->numProcesses);
  for (unsigned int p = 0; p < solver->numProcesses; ++p){
    shiftMoves[p].resize(solver->mr->numMachines);
    for (unsigned int m = 0; m < solver->mr->numMachines; ++m){
      shiftMoves[p][m].process = p;
      shiftMoves[p][m].machine = m;
    }
  }
  // swap
  swapMoves.resize(solver->numServices);
  swapMovesByProcess.resize(solver->numProcesses);
  for (unsigned int s = 0; s < solver->numServices; ++s){
    vector<unsigned int> &processes = solver->services[s].processes;
    unsigned int numProcesses = processes.size();
    swapMoves[s].resize(numProcesses * (numProcesses-1)/2);
    for (unsigned int pp1 = 0, localId = 0; pp1 < numProcesses-1; ++pp1){
      for (unsigned int pp2 = pp1+1; pp2 < numProcesses; ++pp2, ++localId){
        swapMoves[s][localId].process1 = processes[pp1];
        swapMoves[s][localId].process2 = processes[pp2];
        swapMovesByProcess[processes[pp1]].push_back(&swapMoves[s][localId]);
        swapMovesByProcess[processes[pp2]].push_back(&swapMoves[s][localId]);
      }
    }
  }
  // adjustment
  adjustmentsForProcess.resize(solver->numProcesses);
  adjustmentsForNeighborhood.resize(solver->numProcesses);
  for (unsigned int p = 0; p < solver->numProcesses; ++p){
    adjustmentsForProcess[p].id = p;
    adjustments.push_back(&adjustmentsForProcess[p]);

    adjustmentsForNeighborhood[p].resize(solver->mr->numNeighborhoods);
    for (unsigned int n = 0; n < solver->mr->numNeighborhoods; ++n){
      adjustmentsForNeighborhood[p][n].id = n;
      adjustments.push_back(&adjustmentsForNeighborhood[p][n]);
    }
  }
}
    
// ----- initialize tabu search; Calculate Penalty & Diffs from scratch -----
void MachineReassignment::Solver::initialize(Solution &solution){
  solution.clearPenalty();
  int64 constant = con->getConstant();
  solution.penalty.soft += constant;
  solution.eval += constant;

  moves.clear();
  for (unsigned int c = 0; c < con->constraints.size(); ++c){
    Constraint &constraint = *con->constraints[c];
    int64 penalty = constraint.calculatePenalty(solution);
    if (constraint.weight0 == inf){ solution.penalty.hard += penalty; }
    else{ solution.penalty.soft += constraint.weight0 * penalty; }
    solution.eval += constraint.weight * penalty;

    constraint.diffs.clear();
    constraint.calculateDiff(solution);
    for (vector<DiffPenalty>::iterator ite = constraint.diffs.begin(); ite != constraint.diffs.end(); ++ite){
      ite->move->diffEval += constraint.weight * ite->diff;
    }
  }
}

// ----- Apply move; Update Penalty & Diffs -----
void MachineReassignment::Solver::applyMove(Solution &solution, Move *move){
  // --- update tabu list & enumerate constraints to be updated 
  vector<unsigned int> toBeUpdated;
  switch(move->type){
  case Shift: 
    {
      ShiftMove *shift = static_cast<ShiftMove*>(move);
      vector<unsigned int>::iterator in = con->constraintIds[shift->process][shift->machine].begin();
      vector<unsigned int>::iterator out = con->constraintIds[shift->process][solution.assignment[shift->process]].begin();
      vector<unsigned int>::iterator pr = con->constraintIds2[shift->process].begin();
      vector<unsigned int>::iterator mc_in = con->constraintIds3[shift->machine].begin();
      vector<unsigned int>::iterator mc_out = con->constraintIds3[solution.assignment[shift->process]].begin();
      while (1){
        unsigned int c = min(min(min(*in, *out), *pr), min(*mc_in, *mc_out));
        if (c >= con->constraints.size()) break;
        toBeUpdated.push_back(c);
        if (c == *in) ++in;
        if (c == *out) ++out;
        if (c == *pr) ++pr;
        if (c == *mc_in) ++mc_in;
        if (c == *mc_out) ++mc_out;
      }
      break;
    }
  case Swap:
    {
      SwapMove *swap = static_cast<SwapMove*>(move);
      unsigned int process1 = swap->process1;
      unsigned int process2 = swap->process2;
      unsigned int out_value1 = solution.assignment[process1];
      unsigned int out_value2 = solution.assignment[process2];
      vector<unsigned int>::iterator in1 = con->constraintIds[process1][out_value2].begin();
      vector<unsigned int>::iterator in2 = con->constraintIds[process2][out_value1].begin();
      vector<unsigned int>::iterator out1 = con->constraintIds[process1][out_value1].begin();
      vector<unsigned int>::iterator out2 = con->constraintIds[process2][out_value2].begin();
      vector<unsigned int>::iterator pr1 = con->constraintIds2[process1].begin();
      vector<unsigned int>::iterator pr2 = con->constraintIds2[process2].begin();
      vector<unsigned int>::iterator mc1 = con->constraintIds3[out_value1].begin();
      vector<unsigned int>::iterator mc2 = con->constraintIds3[out_value2].begin();
      while (1){
        unsigned int c = min(min(min(*in1, *in2), min(*out1, *out2)), min(min(*pr1, *pr2), min(*mc1, *mc2)));
        if (c >= con->constraints.size()) break;
        toBeUpdated.push_back(c);
        if (c == *in1) ++in1;
        if (c == *in2) ++in2;
        if (c == *out1) ++out1;
        if (c == *out2) ++out2;
        if (c == *pr1) ++pr1;
        if (c == *pr2) ++pr2;
        if (c == *mc1) ++mc1;
        if (c == *mc2) ++mc2;
      }
    break;
    }
  }
  // --- update penalty
  for (vector<unsigned int>::iterator ite = toBeUpdated.begin(); ite != toBeUpdated.end(); ++ite){
    Constraint &constraint = *con->constraints[*ite];
    int64 diff = constraint.updatePenalty(solution, move);
    if (constraint.weight0 == inf){ solution.penalty.hard += diff; }
    else{ solution.penalty.soft += constraint.weight0 * diff; }
    solution.eval += constraint.weight * diff;
  }
  // --- apply move
  move->apply(solution);
  // --- update diff
  for (vector<unsigned int>::iterator ite = toBeUpdated.begin(); ite != toBeUpdated.end(); ++ite){
    Constraint &constraint = *con->constraints[*ite];
    if (constraint.skip) continue;
    vector<DiffPenalty>::iterator dite;
    for (dite = constraint.diffs.begin(); dite != constraint.diffs.end(); ++dite){
      dite->move->diffEval -= constraint.weight * dite->diff;

    }
    constraint.diffs.clear();
    constraint.calculateDiff(solution);
    for (dite = constraint.diffs.begin(); dite != constraint.diffs.end(); ++dite){
      dite->move->diffEval += constraint.weight * dite->diff;
    }
  }
}

// ----- UpdateIncumbent -----
void MachineReassignment::Solver::updateIncumbent(Solution &solution){
  incumbent = solution;
#if !SILENT
  cout << "penalty = " << incumbent.penalty.hard << "/" << incumbent.penalty.soft << " (hard/soft), ";
  cout.setf(ios::fixed); cout.precision(2);
  cout << "time = " << cputime() <<"(s), ";
  cout.unsetf(ios::fixed);
  cout << "iteration = " << iteration << endl;
#endif
}

// ----- Kick -----
void MachineReassignment::Solver::kick(){
  unsigned int numKick = rand(1,3);
  for (unsigned int kick = 0; kick < numKick; ++kick){
    unsigned int process = rand(0, numProcesses-1);
    unsigned int machine = (current.assignment[process]+rand(1,mr->numMachines-1))%mr->numMachines;
    moves.shiftMoves[process][machine].apply(current);
    tabulist.setTabu(process, rand(1,5));
  }
  initialize(current);
}

// ----- Solve -----
bool MachineReassignment::Solver::solve(){ // mr->solution.assignment is used for initial solution
  current.initialize(numProcesses);
  localBest.initialize(numProcesses);
  incumbent.initialize(numProcesses);
  // ----- initial solution
  for (unsigned int p = 0; p < numProcesses; ++p){
    current.assignment[p] = mr->solution.assignment[processes[p].id];
  }
  
  con->prepare(this);
  tabulist.initialize(numProcesses);
  tabulist.setTenure(1);

  moves.initializeMoves(this);

  // ----- 
  initialize(current);
  updateIncumbent(current);
  if (current.penalty == 0) return true;

  do {
    if (useKick){
      kick();
      if (cputime() >= maxCpuTime-0.1) return true;
    }
    if (localSearch()) return true;
    if (!oneGroup) return false;
    useKick = true;
    current = incumbent;
  } while (1);
}

// ----- Local Search: Main Loop -----
bool MachineReassignment::Solver::localSearch(){ // program will be terminated if true is returned
  // for weight adjustment mechanism
  prevBestPenalty = current.penalty;
  numRuns = 0;
  numRounds = 0;
  localBest = current;
  unsigned int s0 = 0, ss = 0;
  while (cputime() < maxCpuTime-0.5){
    double minDiffEval;
    vector<Move*> bestMoves;

    ++iteration;
    minDiffEval = 0;
    bestMoves.clear();
    tabulist.increment();
    
    unsigned int s;
    for (; ss < numServices; ++ss){
      s = (s0 + ss)%numServices;
      for (vector<unsigned int>::iterator pite = services[s].processes.begin(); pite != services[s].processes.end(); ++pite){
        unsigned int p = *pite;
    
        for (vector<ShiftMove>::iterator candidate = moves.shiftMoves[p].begin(); candidate != moves.shiftMoves[p].end(); ++candidate){
            double diffEval = 0;
            unsigned int process = candidate->process;
            unsigned int machine = candidate->machine;
            if (current.assignment[process] == machine) continue;
            if (tabulist.isTabu(candidate->process)) continue;
            diffEval = candidate->diffEval 
                + moves.adjustmentsForProcess[process].diffEval
                + moves.adjustmentsForNeighborhood[process][mr->machines[machine].neighborhood].diffEval
                ;
            if (diffEval < minDiffEval){ 
              minDiffEval = diffEval;
              bestMoves.clear();
            }
            if (diffEval == minDiffEval){ 
              bestMoves.push_back(&(*candidate)); 
            }
          }
      }    
  
      for (vector<SwapMove>::iterator candidate = moves.swapMoves[s].begin(); candidate != moves.swapMoves[s].end(); ++candidate){
        double diffEval = 0;
        unsigned int process1 = candidate->process1;
        unsigned int process2 = candidate->process2;
        if (current.assignment[process1] == current.assignment[process2]) continue;
        diffEval = candidate->diffEval;
        if (diffEval < minDiffEval){ 
          minDiffEval = diffEval;
          bestMoves.clear();
        }
        if (diffEval == minDiffEval){ 
          bestMoves.push_back(&(*candidate)); 
        }
      }
      if (!bestAdmissible) break;
    }

    bool locallyOptimal = true;
    if (minDiffEval < 0){
      Move *bestMove = bestMoves[rand(0, bestMoves.size()-1)];
      double prevEval = current.eval;
      applyMove(current, bestMove);
      if (current.eval < prevEval){ // check that the bestMove improves the current solution
        locallyOptimal = false;
        s0 = (s+1)%numServices;
        if (current.penalty < localBest.penalty){
          localBest = current;
          if (localBest.penalty < incumbent.penalty){
            updateIncumbent(current);
            if (incumbent.penalty == 0) return true;
          }
        }
      }
      else { // due to rounding error
        switch(bestMove->type){
        case Shift:
          {
            ShiftMove *shift = static_cast<ShiftMove*>(bestMove);
            ShiftMove reverse(shift->process, shift->machine0);
            applyMove(current, &reverse);
            break;
          }
        case Swap:
          {
            applyMove(current, bestMove);
            break;
          }
        }
      }
    }
    if (locallyOptimal && ss < numServices){
      locallyOptimal = false;
      ++ss;
    }
    else{
      ss = 0;
    }

    if (locallyOptimal){
      if (!adjustWeight){
        return false;
      }

    // ----- adjust weights -----
      if (localBest.penalty.hard != 0) return false;
      double ub = static_cast<double>(localBest.penalty.soft);
      ++numRuns;
      vector<Constraint*> &constraints = con->constraints;

      double diff, sumSquare;
      vector<unsigned int> constraintsToBeIncreased;
      for (unsigned int trial = 0; trial <= 1; ++trial){
        unsigned int c;
        if (current.eval >= ub * (1 - eps)){
          if (prevBestPenalty == localBest.penalty  // the incumbent solution remains the same
              || ++numRounds >= maxNumRounds){ 
            return false;
          }
          // reset weights
          double maxWeight = current.eval;
          for (c = 0; c < constraints.size(); ++c){
            maxWeight = max(maxWeight, constraints[c]->weight);
          }
          double ratio = current.eval/maxWeight * ratioResetWeights;
          current.eval = 0;
          for (c = 0; c < constraints.size(); ++c){
            constraints[c]->weight *= ratio;
            current.eval += constraints[c]->weight * constraints[c]->penalty;
          }

          for (unsigned int p = 0; p < moves.shiftMoves.size(); ++p){
            for (vector<ShiftMove>::iterator ite = moves.shiftMoves[p].begin(); ite != moves.shiftMoves[p].end(); ++ite){
              ite->diffEval *= ratio;
            }
          }
          for (unsigned int s = 0; s < moves.swapMoves.size(); ++s){
            for (vector<SwapMove>::iterator ite = moves.swapMoves[s].begin(); ite != moves.swapMoves[s].end(); ++ite){
              ite->diffEval *= ratio;
            }
          }
          for (vector<Move*>::iterator ite = moves.adjustments.begin(); ite != moves.adjustments.end(); ++ite){
            (*ite)->diffEval *= ratio;
          }
          numRuns = 0;
          prevBestPenalty = localBest.penalty;
        }
        // increase weights
        diff = ub - current.eval;
        sumSquare = 0;
        for (c = 0; c < constraints.size(); c++){
          if (constraints[c]->penalty > 0 &&
              constraints[c]->weight < static_cast<double>(constraints[c]->weight0)){
            constraintsToBeIncreased.push_back(c);
            sumSquare += static_cast<double>(constraints[c]->penalty) * constraints[c]->penalty;
          }
        }
        if (sumSquare <= 0){ // this error should be due to rounding errors; re-calculate 
          if (trial == 0){
            initialize(current);
            trial = false;
          }
          else{
            return false;
          }
        }
        else break;
      } 
      double multiplier = diff/sumSquare;
      for (vector<unsigned int>::iterator ite = constraintsToBeIncreased.begin(); ite != constraintsToBeIncreased.end(); ++ite){
        Constraint &constraint = *constraints[*ite];
        double prevWeight = constraint.weight;
        constraint.weight = min(static_cast<double>(constraint.weight0), constraint.weight + multiplier * constraint.penalty);
        if (constraint.weight <= 0){
          cout << constraint.weight << " " << multiplier << " " << constraint.penalty << endl;
          throw runtime_error("Negative weight; Overflow?");
        }
        current.eval += (constraint.weight - prevWeight) * constraint.penalty;
        for (vector<DiffPenalty>::iterator dite = constraint.diffs.begin(); dite != constraint.diffs.end(); ++dite){
          dite->move->diffEval += (constraint.weight - prevWeight) * dite->diff;
        }
      }
      tabulist.clear();
    }
  }
  return true;
}